using System;
using System.Text;
using System.IO.Ports;
using System.Threading;
using Microsoft.SPOT;
using GHIElectronics.NETMF.USBHost;

using PS3ControllerUSB;

namespace PS3ControllerBluetooth
{
    public class Bluetooth
    {
        //HCI event flags        
        public enum hci_event
        {
            HCI_FLAG_CMD_COMPLETE = 0x01,
            HCI_FLAG_CMD_STATUS = 0x02,
            HCI_FLAG_CONN_COMPLETE = 0x04,
            HCI_FLAG_DISCONN_COMPLETE = 0x08,
            HCI_FLAG_INCOMING_REQUEST = 0x10,
            HCI_FLAG_REMOTE_NAME = 0x20,
            HCI_FLAG_REMOTE_VERSION = 0x40,
            HCI_FLAG_CHANGE_PACKET_TYPE = 0x80,     
            HCI_FLAG_QOS_SETUP_COMPLETE = 0x100,
        }
        hci_event hci_event_flag;

        //HCI Events        
        const byte EV_COMMAND_COMPLETE = 0x0e;
        const byte EV_COMMAND_STATUS = 0x0f;
        const byte EV_CONNECT_COMPLETE = 0x03;
        const byte EV_DISCONNECT_COMPLETE = 0x05;
        const byte EV_NUM_COMPLETE_PKT = 0x13;
        const byte EV_INCOMING_CONNECT = 0x04;
        const byte EV_ROLE_CHANGED = 0x12;
        const byte EV_REMOTE_NAME_COMPLETE = 0x07;
        const byte EV_REMOTE_VERSION_INFORMATION = 0x0C;
        const byte EV_CONNECTION_PACKET_TYPE_CHANGED = 0x1D;
        const byte EV_QOS_SETUP_COMPLETE = 0x0D;

        //HCI state - used to set up bluetooth dongle
        const byte HCI_INIT_STATE = 0;
        const byte HCI_RESET_STATE = 1;
        const byte HCI_BUFFER_SIZE_STATE = 2;
        const byte HCI_LOCAL_VERSION_STATE = 3;
        const byte HCI_BDADDR_STATE = 4;
        const byte HCI_SCANNING_STATE = 5;
        const byte HCI_CONNECT_IN_STATE = 6;
        const byte HCI_REMOTE_NAME = 7;
        const byte HCI_AUTHENTICATION_ENABLE = 8;//This actually disables authentication       
        const byte HCI_CONNECTED_STATE = 9;
        const byte HCI_REMOTE_VERSION = 10;        
        const byte HCI_CHANGE_PACKET_TYPE = 11;
        const byte HCI_LINK_POLICY_SETTINGS = 12;
        const byte HCI_LINK_SUPERVISION_TIMEOUT = 13;
        const byte HCI_AFH_HOST_CHANNEL_CLASSIFICATION = 14;
        const byte HCI_DISABLE_SCAN = 15;
        const byte HCI_QOS_SETUP = 16;        
        const byte HCI_AFH_HOST_CHANNEL_CLASSIFICATION2 = 17;
        const byte HCI_DONE_STATE = 18;

        byte hci_state = HCI_INIT_STATE;
        
        uint hci_counter; // counter used for bluetooth hci loops        

        //variables filled from HCI event management
        byte hci_command_packets; //how many packets can host send to controller                
        int hci_handle;        
        uint acl_data_packet_length;
        byte sync_data_packet_length;
        uint acl_data_packets;
        uint sync_data_packets;
        byte hci_version;
        uint hci_revision;
        byte lmp_version;
        uint manufacturer_id;
        uint lmp_subversion;
        byte[] my_bdaddr = new byte[6];// bluetooth address stored least significant byte first

        byte dev_link_type;
        byte dev_role;

        byte[] disc_bdaddr = new byte[6];// Device Bluetooth Address
        byte[] disc_class = new byte[3];// Device Class        
        byte[] remote_name = new byte[40];// first 40 chars of name

        bool remote_name_bool;// Indicate if reading remote name
                
        /* Bluetooth L2CAP states for L2CAP_task() */
        const byte L2CAP_EV_WAIT = 0;
        const byte L2CAP_EV_CONTROL_SETUP = 1;
        const byte L2CAP_EV_CONTROL_REQUEST = 2;
        const byte L2CAP_EV_CONTROL_SUCCESS = 3;        

        const byte L2CAP_EV_INTERRUPT_SETUP = 4;
        const byte L2CAP_EV_INTERRUPT_REQUEST = 5;
        const byte L2CAP_EV_INTERRUPT_SUCCESS = 6;

        const byte L2CAP_EV_HID_ENABLE_SIXAXIS = 7;                
        const byte L2CAP_EV_L2CAP_DONE = 8;

        byte l2cap_state = L2CAP_EV_WAIT;

        // Used For Connection Response - Remember to Include High Byte
        const byte PENDING = 0x01;
        const byte SUCCESSFUL = 0x00;

        /* L2CAP event flags */
        public enum l2cap_event
        {                        
            L2CAP_EV_CONTROL_CONNECTION_REQUEST = 0x01,
            L2CAP_EV_CONTROL_CINFIG_REQUEST = 0x02,
            L2CAP_EV_CONTROL_CONFIG_SUCCESS = 0x04,

            L2CAP_EV_INTERRUPT_CONNECTION_REQUEST = 0x08,
            L2CAP_EV_INTERRUPT_CINFIG_REQUEST = 0x10,
            L2CAP_EV_INTERRUPT_CONFIG_SUCCESS = 0x20,                        
        }
        l2cap_event l2cap_event_status;

        /* L2CAP signaling command */        
        const byte L2CAP_CMD_CONNECTION_REQUEST = 0x02;
        const byte L2CAP_CMD_CONNECTION_RESPONSE = 0x03;
        const byte L2CAP_CMD_CONFIG_REQUEST = 0x04;
        const byte L2CAP_CMD_CONFIG_RESPONSE = 0x05;
        const byte L2CAP_CMD_DISCONNECT_REQUEST = 0x06;
        const byte L2CAP_CMD_DISCONNECT_RESPONSE = 0x07;

        /* L2CAP Channels */                               
        byte[] command_scid = new byte[2];// L2CAP local CID for HID_Control                
        byte[] command_dcid = new byte[2] { 0x40, 0x00 };//0x0040        
        byte[] interrupt_scid = new byte[2];// L2CAP local CID for HID_Interrupt        
        byte[] interrupt_dcid = new byte[2] { 0x41, 0x00 };//0x0041

        /* Bluetooth L2CAP PSM */
        const byte L2CAP_PSM_HID_CTRL = 0x11;// HID_Control        
        const byte L2CAP_PSM_HID_INTR = 0x13;// HID_Interrupt
                      
        //Bluetooth    
        public static byte[] BDaddr = new byte[6] { 0x00, 0x15, 0x83, 0x15, 0xA3, 0x10 };//My bluetooth dongle's address
        public static ushort CSR_VENDOR_ID = 0x0A12;//Cambridge Silicon Radio Ltd.
        public static ushort CSR_PRODUCT_ID = 0x001;//Bluetooth dongle
        public static bool PS3BTConnected;

        private USBH_RawDevice raw;
        private USBH_RawDevice.Pipe IntInPipe, BulkInPipe, BulkOutPipe;

        private byte[] IntInBuffer = new byte[16];
        private byte[] HCIBuffer = new byte[64];
        private byte[] BulkInBuffer = new byte[64];        
        private byte[] L2CAPBuffer = new byte[64];
        private byte[] HIDBuffer = new byte[64];
        
        Thread IntInThread;
        Thread BulkInThread;

        //Setup serial connection
        static SerialPort UART = new SerialPort("COM2", 115200);

        public Bluetooth(USBH_Device device)
        {
            if (device.VENDOR_ID != CSR_VENDOR_ID || device.PRODUCT_ID != CSR_PRODUCT_ID)
                throw new InvalidOperationException();

            raw = new USBH_RawDevice(device);
            USBH_Descriptors.Configuration cd = raw.GetConfigurationDescriptors(0);

            IntInPipe = raw.OpenPipe(cd.interfaces[0].endpoints[0]); // Interrupt In
            BulkInPipe = raw.OpenPipe(cd.interfaces[0].endpoints[1]); // Bulk In
            BulkOutPipe = raw.OpenPipe(cd.interfaces[0].endpoints[2]); // Bulk Out

            IntInPipe.TransferTimeout = 0;
            BulkInPipe.TransferTimeout = 0;

            //Set configuration
            raw.SendSetupTransfer(0x00, 0x09, cd.bConfigurationValue, 0x00);

            for (int i = 0; i < PS3Controller.OUTPUT_REPORT_BUFFER.Length; i++)
                HIDBuffer[i + 2] = PS3Controller.OUTPUT_REPORT_BUFFER[i];//First two bytes reserved for report type and ID

            HIDBuffer[0] = 0x52;// HID BT Set_report (0x50) | Report Output (Input 0x02)
            HIDBuffer[1] = 0x01;// Report ID

            IntInThread = new Thread(IntReadingThread);            
            IntInThread.Start();

            BulkInThread = new Thread(BulkReadingThread);
            BulkInThread.Priority = ThreadPriority.Highest;
            BulkInThread.Start();

            WriteSerial("CSR Initialized");
        }
        public void Abort()
        {
            IntInThread.Abort();
            BulkInThread.Abort();
        }

        private void IntReadingThread()
        {
            while (true)
            {
                //Read every bInterval
                Thread.Sleep(IntInPipe.PipeEndpoint.bInterval);
                {
                    try
                    {
                        IntInPipe.TransferData(IntInBuffer, 0, IntInBuffer.Length);
                        /*
                        // uncomment for debugging
                            int i = 0;
                            WriteSerial(
                                ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " +
                                ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " +
                                ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " +
                                ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++]) + " " + ByteToHex(IntInBuffer[i++])
                                );
                        */
                        switch (IntInBuffer[0])
                        {
                            /*  buf[0] = Event Code                            */
                            /*  buf[1] = Parameter Total Length                */
                            /*  buf[n] = Event Parameters based on each event  */
                            case EV_QOS_SETUP_COMPLETE:
                                if (IntInBuffer[2] == 0x00)//Success                                
                                    hci_event_flag |= hci_event.HCI_FLAG_QOS_SETUP_COMPLETE;
                                break;
                            case EV_CONNECTION_PACKET_TYPE_CHANGED:
                                if (IntInBuffer[2] == 0x00)//Success
                                {
                                    hci_event_flag |= hci_event.HCI_FLAG_CHANGE_PACKET_TYPE;
                                    WriteSerial("Connection Packet Type Changed - Connection Handle: 0x" + ByteToHex(IntInBuffer[4]) + ByteToHex(IntInBuffer[3]) + " Packet Type: 0x" + ByteToHex(IntInBuffer[6]) + ByteToHex(IntInBuffer[5]));                                                                        
                                }
                                break;
                            case EV_REMOTE_VERSION_INFORMATION:
                                if (IntInBuffer[2] == 0x00)//Success
                                {
                                    hci_event_flag |= hci_event.HCI_FLAG_REMOTE_VERSION;//Set flag                                    
                                    WriteSerial("Remote Version Information - Connection Handle: 0x" + ByteToHex(IntInBuffer[4]) + ByteToHex(IntInBuffer[3]) + " LMP Version: 0x" + ByteToHex(IntInBuffer[5]) + " Manufactor Name: 0x" + ByteToHex(IntInBuffer[7]) + ByteToHex(IntInBuffer[6]) + " LMP SubVersion: 0x" + ByteToHex(IntInBuffer[9]) + ByteToHex(IntInBuffer[8]));
                                }
                                break;

                            case EV_REMOTE_NAME_COMPLETE:                                                                                                                             
                                for (byte i = 9; i < IntInBuffer.Length; i++)
                                    remote_name[i - 9] = IntInBuffer[i];
                                remote_name_bool = true;                                
                                break;                        

                            case EV_COMMAND_COMPLETE:
                                //Debug.Print("Command Complete");
                                hci_command_packets = IntInBuffer[2];//Update flow control    
                                if (IntInBuffer[5] != 0x00)// Check if success
                                {
                                    WriteSerial("HCI Command Complete Failed - OGF: 0x" + (IntInBuffer[4] >> 2) + " OCF: 0x" + IntInBuffer[3]);
                                    return;
                                }
                                
                                hci_event_flag |= hci_event.HCI_FLAG_CMD_COMPLETE;//Set flag

                                // parameter from Set AFH Host Channel Classification
                                if ((IntInBuffer[3] == 0x3F) && (IntInBuffer[4] == 0x0C))
                                    Debug.Print("Set AFH Host Channel Classification Complete");
                                
                                // parameter from Write Link Supervision Timeout
                                if ((IntInBuffer[3] == 0x37) && (IntInBuffer[4] == 0x0C))
                                    Debug.Print("Write Link Supervision Timeout Complete");                                                                  
                                
                                // parameters from Write Link Policy Settings
                                if ((IntInBuffer[3] == 0x0D) && (IntInBuffer[4] == 0x08))                                                                   
                                    Debug.Print("Write Link Policy Settings Complete");
                                
                                // parameters from Write Authentication Enable
                                if ((IntInBuffer[3] == 0x20) && (IntInBuffer[4] == 0x0C))
                                    Debug.Print("Authentication Enable Complete");                       
                                
                                // parameters from read local bluetooth address
                                if ((IntInBuffer[3] == 0x09) && (IntInBuffer[4] == 0x10))
                                {
                                    for (byte i = 0; i < 6; i++)
                                        my_bdaddr[i] = IntInBuffer[6 + i];
                                    //Debug.Print("Read Local Bluetooth Address: " + ByteToHex(my_bdaddr[5]).ToString() + ":" + ByteToHex(my_bdaddr[4]).ToString() + ":" + ByteToHex(my_bdaddr[3]).ToString() + ":" + ByteToHex(my_bdaddr[2]).ToString() + ":" + ByteToHex(my_bdaddr[1]).ToString() + ":" + ByteToHex(my_bdaddr[0]).ToString());
                                }

                                //parameters from read buffer size
                                if ((IntInBuffer[3] == 0x05) && (IntInBuffer[4] == 0x10))
                                {
                                    Debug.Print("Read Buffer Size");
                                    acl_data_packet_length = (uint)(IntInBuffer[6] | IntInBuffer[7] << 8);
                                    sync_data_packet_length = IntInBuffer[8];
                                    acl_data_packets = (uint)(IntInBuffer[9] | IntInBuffer[10] << 8);
                                    sync_data_packets = (uint)(IntInBuffer[11] | IntInBuffer[12] << 8);
                                }

                                // parameters from read local version
                                if ((IntInBuffer[3] == 0x01) && (IntInBuffer[4] == 0x10))
                                {
                                    Debug.Print("Read Local Version");
                                    hci_version = IntInBuffer[6];
                                    hci_revision = (uint)(IntInBuffer[7] | IntInBuffer[8] << 8);
                                    lmp_version = IntInBuffer[9];
                                    manufacturer_id = (uint)(IntInBuffer[10] | IntInBuffer[11] << 8);
                                    lmp_subversion = (uint)(IntInBuffer[12] | IntInBuffer[13] << 8);
                                }

                                break;

                            case EV_COMMAND_STATUS:
                                //Debug.Print("Command Status");
                                hci_command_packets = IntInBuffer[3];// update flow control
                                hci_event_flag |= hci_event.HCI_FLAG_CMD_STATUS;

                                if (IntInBuffer[2] == 0)                                    
                                    Debug.Print("Command Status - Complete: 0x" + ByteToHex(IntInBuffer[5]) + ByteToHex(IntInBuffer[4]));
                                else// show status on serial if not OK
                                    WriteSerial("HCI Command Failed - Error: 0x" + ByteToHex(IntInBuffer[2]) + " Command: 0x" + ByteToHex(IntInBuffer[5]) + ByteToHex(IntInBuffer[4]));
                                break;

                            case EV_CONNECT_COMPLETE:
                                if (IntInBuffer[2] == 0)// check if connected OK
                                {                                    
                                    hci_handle = IntInBuffer[3] | IntInBuffer[4] << 8; //store the handle for the ACL connection
                                    //Debug.Print("Connect Complete: " + hci_handle);
                                    hci_event_flag |= hci_event.HCI_FLAG_CONN_COMPLETE; //set connection OK flag                                                                          
                                }
                                break;

                            case EV_DISCONNECT_COMPLETE:
                                if (IntInBuffer[2] == 0)// check if disconnected OK
                                {                                    
                                    string reason = "Unknown Reason";
                                    if (IntInBuffer[5] == 0x08)
                                        reason = "Connection Timeout";
                                    else if (IntInBuffer[5] == 0x13)
                                        reason = "Remote User Terminated Connection";
                                    else if (IntInBuffer[5] == 0x14)
                                        reason = "Remote Device Terminated Connection due to Low Resources";
                                    else if (IntInBuffer[5] == 0x15)
                                        reason = "Remote Device Terminated Connection due to Power Off";
                                    else if (IntInBuffer[5] == 0x16)
                                        reason = "Connection Terminated By Local Host";

                                    WriteSerial("Disconnect Complete - Reason: " + reason);

                                    hci_event_flag |= hci_event.HCI_FLAG_DISCONN_COMPLETE;
                                    hci_event_flag &= ~(hci_event.HCI_FLAG_CONN_COMPLETE); //clear connection OK flag                                    
                                }
                                break;

                            case EV_NUM_COMPLETE_PKT:                                                                
                                Debug.Print("Number Of Completed Packets: " + (uint)(IntInBuffer[6] | IntInBuffer[7] << 8));
                                break;

                            case EV_INCOMING_CONNECT:
                                disc_bdaddr[0] = IntInBuffer[2];
                                disc_bdaddr[1] = IntInBuffer[3];
                                disc_bdaddr[2] = IntInBuffer[4];
                                disc_bdaddr[3] = IntInBuffer[5];
                                disc_bdaddr[4] = IntInBuffer[6];
                                disc_bdaddr[5] = IntInBuffer[7];
                                disc_class[0] = IntInBuffer[8];
                                disc_class[1] = IntInBuffer[9];
                                disc_class[2] = IntInBuffer[10];
                                dev_link_type = IntInBuffer[11];

                                Debug.Print("Incoming Connect");
                                Debug.Print("Bluetooth Address: " + ByteToHex(disc_bdaddr[5]).ToString() + ":" + ByteToHex(disc_bdaddr[4]).ToString() + ":" + ByteToHex(disc_bdaddr[3]).ToString() + ":" + ByteToHex(disc_bdaddr[2]).ToString() + ":" + ByteToHex(disc_bdaddr[1]).ToString() + ":" + ByteToHex(disc_bdaddr[0]).ToString());
                                if (disc_class[2] == 0 && disc_class[1] == 5 && disc_class[0] == 8)//Major Sevice Class = None (0x00), Major Device Class = Peripheral (0x05), Minor Device Class Field = Gamepad (0x08)
                                    Debug.Print("Class - Major Device Class: Peripheral (0x05), Minor Device Class Field: Gamepad (0x08)");
                                else
                                    Debug.Print("Class: " + ByteToHex(disc_class[2]) + " " + ByteToHex(disc_class[1]) + " " + ByteToHex(disc_class[0]));

                                if (dev_link_type == 0x00)//0 = SCO
                                    Debug.Print("Link Type: SCO");
                                else if (dev_link_type == 0x01)//1 = ACL
                                    Debug.Print("Link Type: ACL");
                                else if (dev_link_type == 0x02)//2 = eSCO
                                    Debug.Print("Link Type: eSCO");
                                
                                hci_event_flag |= hci_event.HCI_FLAG_INCOMING_REQUEST;
                                break;

                            case EV_ROLE_CHANGED:
                                Debug.Print("Role Changed");
                                dev_role = IntInBuffer[9];
                                break;

                            default:
                                if (remote_name_bool)
                                {                                    
                                    for (byte i2 = 0; i2 < IntInBuffer.Length; i2++)
                                    {
                                        if (IntInBuffer[i2] == 0x00)
                                        {
                                            Debug.Print("Finished Reading Name");
                                            remote_name[38] = 13;// Carriage Return
                                            remote_name[39] = 10;// Line Feed                                                                                                                                                                                                                    
                                            remote_name_bool = false;
                                            hci_event_flag |= hci_event.HCI_FLAG_REMOTE_NAME;
                                            break;
                                        }
                                        if (IntInBuffer[0] == 0x54)// Second Line
                                            remote_name[7 + i2] = IntInBuffer[i2];
                                        else if (IntInBuffer[0] == 0x6C)// Third Line
                                            remote_name[23 + i2] = IntInBuffer[i2];
                                        else
                                            break;
                                    }                                    
                                }
                                else if (IntInBuffer[0] != 0x00)
                                    Debug.Print("Unmanaged event: " + ByteToHex(IntInBuffer[0]));
                                break;
                        }
                        HCI_task();//Start HCI_task
                    }
                    catch (Exception ex)
                    {
                        Debug.Print("==============================");
                        Debug.Print(DateTime.Now.ToString());
                        Debug.Print(ex.Message);
                        Debug.Print(ex.StackTrace);
                        if (ex.InnerException != null)
                        {
                            Debug.Print("Inner Exception: " + ex.InnerException.Message);
                        }
                        Thread.Sleep(100);
                    }
                }
            }
        }
        void HCI_task()
        {
            switch (hci_state)
            {
                case HCI_INIT_STATE:
                    hci_counter++;
                    if (hci_counter > 10)
                    {
                        // wait until we have looped 10 times to clear any old events
                        WriteSerial("Init State");
                        hci_reset();
                        hci_state = HCI_RESET_STATE;
                        hci_counter = 0;
                    }
                    break;

                case HCI_RESET_STATE:
                    hci_counter++;
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("HCI Reset Complete");
                        hci_state = HCI_BUFFER_SIZE_STATE;
                        hci_read_buffer_size();
                    }
                    if (hci_counter > 100)
                    {
                        WriteSerial("No Response to HCI Reset - Try reconnecting the Bluetooth Dongle");
                        hci_state = HCI_INIT_STATE;
                        hci_counter = 0;
                    }
                    break;

                case HCI_BUFFER_SIZE_STATE:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("ACL Data Packet Length: " + acl_data_packet_length);
                        WriteSerial("SCO Data Packet Length: " + sync_data_packet_length);
                        WriteSerial("Total ACL Data Packets: " + acl_data_packets);
                        WriteSerial("Total SCO Data Packets: " + sync_data_packets);
                        hci_state = HCI_LOCAL_VERSION_STATE;
                        hci_read_local_version();
                    }
                    break;

                case HCI_LOCAL_VERSION_STATE:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("HCI Version: " + hci_version);
                        WriteSerial("HCI Revision: " + hci_revision);
                        WriteSerial("LMP Version: " + lmp_version);
                        WriteSerial("Manufactor ID: " + manufacturer_id);
                        WriteSerial("LMP Subversion: " + lmp_subversion);
                        hci_state = HCI_BDADDR_STATE;
                        hci_read_bdaddr();
                    }
                    break;

                case HCI_BDADDR_STATE:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Local Bluetooth Address: " + ByteToHex(my_bdaddr[5]).ToString() + ":" + ByteToHex(my_bdaddr[4]).ToString() + ":" + ByteToHex(my_bdaddr[3]).ToString() + ":" + ByteToHex(my_bdaddr[2]).ToString() + ":" + ByteToHex(my_bdaddr[1]).ToString() + ":" + ByteToHex(my_bdaddr[0]).ToString());
                        BDaddr[0] = my_bdaddr[5];
                        BDaddr[1] = my_bdaddr[4];
                        BDaddr[2] = my_bdaddr[3];
                        BDaddr[3] = my_bdaddr[2];
                        BDaddr[4] = my_bdaddr[1];
                        BDaddr[5] = my_bdaddr[0];

                        hci_state = HCI_SCANNING_STATE;
                    }
                    break;

                case HCI_SCANNING_STATE:
                    WriteSerial("Wait For Incoming Connect Request");
                    hci_write_scan_enable();
                    hci_state = HCI_CONNECT_IN_STATE;
                    break;

                case HCI_CONNECT_IN_STATE:
                    if (hciflag(hci_event.HCI_FLAG_INCOMING_REQUEST))
                    {
                        WriteSerial("Incoming Request");                                                                     
                        hci_remote_name(0);
                        hci_state = HCI_REMOTE_NAME;                                             
                    }
                    break;
                                  
                case HCI_REMOTE_NAME:
                    if (hciflag(hci_event.HCI_FLAG_REMOTE_NAME))
                    {
                        WriteSerial("Remote Name:");
                        UART.Write(remote_name, 0, 40);

                        hci_authentication_enable();
                        hci_state = HCI_AUTHENTICATION_ENABLE;
                    }
                    break;                
            case HCI_AUTHENTICATION_ENABLE:
                if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                {
                    WriteSerial("Authentication Enable Complete");                    
                    hci_accept_connection(0);
                    hci_state = HCI_CONNECTED_STATE;
                }
                break;
                           
            case HCI_CONNECTED_STATE:
                if (hciflag(hci_event.HCI_FLAG_CONN_COMPLETE))
                {
                    WriteSerial("Connected to Device: " + ByteToHex(disc_bdaddr[5]) + ":" + ByteToHex(disc_bdaddr[4]) + ":" + ByteToHex(disc_bdaddr[3]) + ":" + ByteToHex(disc_bdaddr[2]) + ":" + ByteToHex(disc_bdaddr[1]) + ":" + ByteToHex(disc_bdaddr[0]));

                    hci_remote_version();
                    hci_state = HCI_REMOTE_VERSION;
                }
                break;
                    
                case HCI_REMOTE_VERSION:
                    if (hciflag(hci_event.HCI_FLAG_REMOTE_VERSION))
                    {
                        hci_change_packet_type();
                        hci_state = HCI_CHANGE_PACKET_TYPE;
                    }
                    break;                                 
                case HCI_CHANGE_PACKET_TYPE:
                    if (hciflag(hci_event.HCI_FLAG_CHANGE_PACKET_TYPE))
                    {
                        hci_link_policy_settings();
                        hci_state = HCI_LINK_POLICY_SETTINGS;
                    }
                    break;
                case HCI_LINK_POLICY_SETTINGS:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Write Link Policy Settings Complete");                        
                        hci_link_supervision_timeout();
                        hci_state = HCI_LINK_SUPERVISION_TIMEOUT;
                    }
                    break;
                case HCI_LINK_SUPERVISION_TIMEOUT:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Write Link Supervision Timeout Complete");                        
                        hci_afh_host_channel_classification();
                        hci_state = HCI_AFH_HOST_CHANNEL_CLASSIFICATION;                        
                    }
                    break;

                case HCI_AFH_HOST_CHANNEL_CLASSIFICATION:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Set AFH Host Channel Classification Complete");
                        hci_write_scan_disable();//Only allow one controller - good for debugging 
                        hci_state = HCI_DISABLE_SCAN;                 
                    }
                    break;
                case HCI_DISABLE_SCAN:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Scan Disabled");
                        hci_qos_setup();
                        hci_state = HCI_QOS_SETUP;
                    }
                    break;

                case HCI_QOS_SETUP:
                    if (hciflag(hci_event.HCI_FLAG_QOS_SETUP_COMPLETE))
                    {
                        WriteSerial("QOS Setup Complete");
                        hci_afh_host_channel_classification();
                        hci_state = HCI_AFH_HOST_CHANNEL_CLASSIFICATION2;
                    }
                    break;

                case HCI_AFH_HOST_CHANNEL_CLASSIFICATION2:
                    if (hciflag(hci_event.HCI_FLAG_CMD_COMPLETE))
                    {
                        WriteSerial("Set AFH Host Channel Classification Complete");
                        WriteSerial("HCI Done");
                        l2cap_state = L2CAP_EV_CONTROL_SETUP;
                        hci_state = HCI_DONE_STATE;
                    }
                    break;                    

                case HCI_DONE_STATE:
                    if (hciflag(hci_event.HCI_FLAG_DISCONN_COMPLETE))
                    {
                        disconnectPS3Controller();
                        WriteSerial("Disconnected from Device: " + ByteToHex(disc_bdaddr[5]) + ":" + ByteToHex(disc_bdaddr[4]) + ":" + ByteToHex(disc_bdaddr[3]) + ":" + ByteToHex(disc_bdaddr[2]) + ":" + ByteToHex(disc_bdaddr[1]) + ":" + ByteToHex(disc_bdaddr[0]));
                        PS3BTConnected = false;
                        l2cap_state = L2CAP_EV_WAIT;
                        hci_state = HCI_SCANNING_STATE;
                    }
                    break;                                     

                default:
                    break;                               

            }
        }

        private void BulkReadingThread()
        {
            while (true)
            {
                //Read every bInterval
                Thread.Sleep(BulkInPipe.PipeEndpoint.bInterval);                
                {
                    try
                    {
                        BulkInPipe.TransferData(BulkInBuffer, 0, BulkInBuffer.Length);

                        // uncomment for debugging
                        /*
                        int i = 0;
                        WriteSerial(
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +
                            ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++])
                            );
                        */

                        if (((BulkInBuffer[0] | (BulkInBuffer[1] << 8)) == (hci_handle | 0x2000)))//acl_handle_ok
                        {
                            if ((BulkInBuffer[6] | (BulkInBuffer[7] << 8)) == 0x0001)//l2cap_control - Channel ID for ACL-U                                
                            {
                                if (BulkInBuffer[8] != 0x00)
                                    Debug.Print("L2CAP Signaling Command - 0x" + ByteToHex(BulkInBuffer[8]));

                                if (BulkInBuffer[8] == L2CAP_CMD_CONNECTION_REQUEST)
                                {
                                    //Debug.Print("Code: 0x" + ByteToHex(BulkInBuffer[8]) + " Identifier: 0x" + ByteToHex(BulkInBuffer[9]) + " Length: 0x" + ByteToHex(BulkInBuffer[11]) + ByteToHex(BulkInBuffer[10]) + " PSM: 0x" + ByteToHex(BulkInBuffer[13]) + ByteToHex(BulkInBuffer[12]) + " SCID: 0x" + ByteToHex(BulkInBuffer[15]) + ByteToHex(BulkInBuffer[14]));                                    
                                    Debug.Print("L2CAP Connection Request - Source CID: 0x" + ByteToHex(command_scid[1]) + ByteToHex(command_scid[0]) + " PSM: 0x" + ByteToHex(BulkInBuffer[13]) + ByteToHex(BulkInBuffer[12]));
                                    if ((BulkInBuffer[13] | BulkInBuffer[12]) == L2CAP_PSM_HID_CTRL)
                                    {
                                        command_scid[0] = BulkInBuffer[14];
                                        command_scid[1] = BulkInBuffer[15];
                                        l2cap_event_status |= l2cap_event.L2CAP_EV_CONTROL_CONNECTION_REQUEST;
                                    }
                                    if ((BulkInBuffer[13] | BulkInBuffer[12]) == L2CAP_PSM_HID_INTR)
                                    {
                                        interrupt_scid[0] = BulkInBuffer[14];
                                        interrupt_scid[1] = BulkInBuffer[15];
                                        l2cap_event_status |= l2cap_event.L2CAP_EV_INTERRUPT_CONNECTION_REQUEST;
                                    }
                                }
                                else if (BulkInBuffer[8] == L2CAP_CMD_CONFIG_RESPONSE)
                                {
                                    if (BulkInBuffer[12] == command_dcid[0] && BulkInBuffer[13] == command_dcid[1])
                                    {
                                        if ((BulkInBuffer[16] | (BulkInBuffer[17] << 8)) == 0x0000)//Success
                                        {
                                            Debug.Print("HID Control Configuration Complete");
                                            l2cap_event_status |= l2cap_event.L2CAP_EV_CONTROL_CONFIG_SUCCESS;
                                        }
                                    }
                                    else if (BulkInBuffer[12] == interrupt_dcid[0] && BulkInBuffer[13] == interrupt_dcid[1])
                                    {
                                        if ((BulkInBuffer[16] | (BulkInBuffer[17] << 8)) == 0x0000)//Success
                                        {
                                            Debug.Print("HID Interrupt Configuration Complete");
                                            l2cap_event_status |= l2cap_event.L2CAP_EV_INTERRUPT_CONFIG_SUCCESS;
                                        }
                                    }
                                }
                                else if (BulkInBuffer[8] == L2CAP_CMD_CONFIG_REQUEST)
                                {
                                    if (BulkInBuffer[12] == command_dcid[0] && BulkInBuffer[13] == command_dcid[1])
                                    {
                                        Debug.Print("HID Control Configuration Request");
                                        l2cap_event_status |= l2cap_event.L2CAP_EV_CONTROL_CINFIG_REQUEST;
                                    }
                                    else if (BulkInBuffer[12] == interrupt_dcid[0] && BulkInBuffer[13] == interrupt_dcid[1])
                                    {
                                        Debug.Print("HID Interrupt Configuration Request");
                                        l2cap_event_status |= l2cap_event.L2CAP_EV_INTERRUPT_CINFIG_REQUEST;
                                    }
                                }                                    
                                else if (BulkInBuffer[8] == L2CAP_CMD_DISCONNECT_REQUEST)//l2cap_disconnect_request
                                {
                                    if (BulkInBuffer[12] == command_dcid[0] && BulkInBuffer[13] == command_dcid[1])                                    
                                        WriteSerial("Disconnected from Device - Control");

                                    else if (BulkInBuffer[12] == interrupt_dcid[0] && BulkInBuffer[13] == interrupt_dcid[1])
                                        WriteSerial("Disconnected from Device - Interrupt");                                                                                
                                }                                     
                                L2CAP_task();
                            }
                                
                            else if (BulkInBuffer[6] == interrupt_dcid[0] && BulkInBuffer[7] == interrupt_dcid[1])//l2cap_interrupt
                            {                                
                                //Debug.Print("L2CAP Interrupt");          
                                PS3BTConnected = true;
                                //readReport();//Uncomment for debugging
                            }                          
                        }

                    }
                    catch (Exception ex)
                    {
                        Debug.Print("==============================");
                        Debug.Print(DateTime.Now.ToString());
                        Debug.Print(ex.Message);
                        Debug.Print(ex.StackTrace);
                        if (ex.InnerException != null)
                        {
                            Debug.Print("Inner Exception: " + ex.InnerException.Message);
                        }
                        Thread.Sleep(100);
                    }
                }
            }
        }
        void L2CAP_task()
        {
            switch (l2cap_state)
            {
                case L2CAP_EV_WAIT:
                    break;
                case L2CAP_EV_CONTROL_SETUP:
                    if (l2capflag(l2cap_event.L2CAP_EV_CONTROL_CONNECTION_REQUEST))
                    {                        
                        WriteSerial("HID Control Incoming Connection Request");                                                
                        l2cap_connection_response(0x01, command_dcid, command_scid, PENDING);                        
                        l2cap_connection_response(0x01, command_dcid, command_scid, SUCCESSFUL);                                                
                        l2cap_configure_request(0x02, command_scid);

                        l2cap_state = L2CAP_EV_CONTROL_REQUEST;
                    }
                    break;
                case L2CAP_EV_CONTROL_REQUEST:
                    if (l2capflag(l2cap_event.L2CAP_EV_CONTROL_CINFIG_REQUEST))
                    {
                        WriteSerial("HID Control Configuration Request");
                        l2cap_config_response(0x02, command_scid);
                        l2cap_state = L2CAP_EV_CONTROL_SUCCESS;                        
                    }
                    break;

                case L2CAP_EV_CONTROL_SUCCESS:
                    if (l2capflag(l2cap_event.L2CAP_EV_CONTROL_CONFIG_SUCCESS))
                    {
                        WriteSerial("HID Control Successfully Configured");
                        l2cap_state = L2CAP_EV_INTERRUPT_SETUP;
                    }
                    break;                    
                case L2CAP_EV_INTERRUPT_SETUP:
                    if (l2capflag(l2cap_event.L2CAP_EV_INTERRUPT_CONNECTION_REQUEST))
                    {
                        WriteSerial("HID Interrupt Incoming Connection Request");                        
                        l2cap_connection_response(0x03, interrupt_dcid, interrupt_scid, PENDING);
                        l2cap_connection_response(0x03, interrupt_dcid, interrupt_scid, SUCCESSFUL);                        
                        l2cap_configure_request(0x0B, interrupt_scid);                        

                        l2cap_state = L2CAP_EV_INTERRUPT_REQUEST;
                    }
                    break;
                case L2CAP_EV_INTERRUPT_REQUEST:
                    if (l2capflag(l2cap_event.L2CAP_EV_INTERRUPT_CINFIG_REQUEST))
                    {
                        WriteSerial("HID Interrupt Configuration Request");
                        l2cap_config_response(0x04, interrupt_scid);
                        l2cap_state = L2CAP_EV_INTERRUPT_SUCCESS;                        
                    }
                    break;                    
                case L2CAP_EV_INTERRUPT_SUCCESS:
                    if (l2capflag(l2cap_event.L2CAP_EV_INTERRUPT_CONFIG_SUCCESS))
                    {
                        WriteSerial("HID Interrupt Successfully Configured");
                        l2cap_state = L2CAP_EV_HID_ENABLE_SIXAXIS;
                    }
                    break;
                case L2CAP_EV_HID_ENABLE_SIXAXIS:
                    Thread.Sleep(1000);
                    hid_enable_sixaxis();
                    WriteSerial("Sixaxis Enabled");
                    hid_setLedOn(PS3Controller.LED.LED1);                    

                    WriteSerial("HID Done");
                    l2cap_state = L2CAP_EV_L2CAP_DONE;
                    break;
                case L2CAP_EV_L2CAP_DONE:                    
                    break;

            }
        }
        public enum Button
        {
            // byte location | bit location
            SELECT = (11 << 8) | 0x01,
            L3 = (11 << 8) | 0x02,
            R3 = (11 << 8) | 0x04,
            START = (11 << 8) | 0x08,
            UP = (11 << 8) | 0x10,
            RIGHT = (11 << 8) | 0x20,
            DOWN = (11 << 8) | 0x40,
            LEFT = (11 << 8) | 0x80,

            L2 = (12 << 8) | 0x01,
            R2 = (12 << 8) | 0x02,
            L1 = (12 << 8) | 0x04,
            R1 = (12 << 8) | 0x08,
            TRIANGLE = (12 << 8) | 0x10,
            CIRCLE = (12 << 8) | 0x20,
            CROSS = (12 << 8) | 0x40,
            SQUARE = (12 << 8) | 0x80,

            PS = (13 << 8) | 0x01,
        }
        public enum AnalogHat
        {
            LeftHatX = 15,
            LeftHatY = 16,
            RightHatX = 17,
            RightHatY = 18,
        }
        public bool GetButton(Button b)
        {            
            if (BulkInBuffer == null)
                return false;                        
            if ((BulkInBuffer[(uint)b >> 8] & ((byte)b & 0xff)) > 0)
                    return true;
            else
                return false;
        }
        public byte GetAnalogHat(AnalogHat a)
        {
            if (BulkInBuffer == null)
                return 0;                        
            return (byte)(BulkInBuffer[(uint)a]);            
        }
        void disconnectPS3Controller()
        {
            l2cap_disconnection_request(0x1C, interrupt_dcid, interrupt_scid);
            l2cap_disconnection_request(0x1D, command_dcid, command_scid);            
            hci_disconnect();
        }
        /************************************************************/
        /*             HID Report (HCI ACL Packet)                  */
        /************************************************************/
        void readReport()
        {
            if (BulkInBuffer[8] == 0xA1)//HID_THDR_DATA_INPUT
            {                
                /*
                int i = 11;
                WriteSerial(
                    ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " +                   
                    ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++]) + " " + ByteToHex(BulkInBuffer[i++])                                                          
                    );                
                 */
            }
        }

        /*                          HCI ACL Data Packet
        *
        *   buf[0]          buf[1]          buf[2]          buf[3]
        *   0       4       8    11 12      16              24            31 MSB
        *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
        *  |      HCI Handle       |PB |BC |       Data Total Length       |   HCI ACL Data Packet
        *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
        *
        *   buf[4]          buf[5]          buf[6]          buf[7]
        *   0               8               16                            31 MSB
        *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
        *  |            Length             |          Channel ID           |   Basic L2CAP header
        *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
        *
        *   buf[8]          buf[9]          buf[10]         buf[11]
        *   0               8               16                            31 MSB
        *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
        *  |     Code      |  Identifier   |            Length             |   Control frame (C-frame)
        *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.   (signaling packet format)
        */
        /************************************************************/
        /*                    HID Commands                          */
        /************************************************************/
        void HID_Command(byte[] data, int length)
        {
            byte[] buf = new byte[64];
            buf[0] = (byte)(hci_handle & 0xff);    // HCI handle with PB,BC flag
            buf[1] = (byte)(((hci_handle >> 8) & 0x0f) | 0x20);
            buf[2] = (byte)((4 + length) & 0xff); // HCI ACL total data length
            buf[3] = (byte)((4 + length) >> 8);
            buf[4] = (byte)(length & 0xff); // L2CAP header: Length
            buf[5] = (byte)(length >> 8);
            buf[6] = command_scid[0];
            buf[7] = command_scid[1];

            for (uint i = 0; i < length; i++)//L2CAP C-frame            
                buf[8 + i] = data[i];

            BulkOutPipe.TransferData(buf, 0, length + 8);
            Thread.Sleep(150);//There have to be a delay between commands
        }
        void hid_setAllOff()
        {
            for (int i = 0; i < PS3Controller.OUTPUT_REPORT_BUFFER.Length; i++)
                HIDBuffer[i + 2] = PS3Controller.OUTPUT_REPORT_BUFFER[i];//First two bytes reserved for report type and ID

            HID_Command(HIDBuffer, PS3Controller.OUTPUT_REPORT_BUFFER.Length + 2);
        }
        void hid_setRumbleOff()
        {
            HIDBuffer[3] = 0xfe;
            HIDBuffer[4] = 0x00;//low mode off
            HIDBuffer[5] = 0xfe;
            HIDBuffer[6] = 0x00;//high mode off

            HID_Command(HIDBuffer, PS3Controller.OUTPUT_REPORT_BUFFER.Length + 2);
        }
        void hid_setRumbleOn(PS3Controller.Rumble a)
        {
            if (((uint)a & 0x30) > 0)
            {
                HIDBuffer[3] = 0xfe;
                HIDBuffer[5] = 0xfe;
                if (((uint)a & 0x10) > 0)
                {
                    HIDBuffer[4] = 0;//low mode off
                    HIDBuffer[6] = 0xff;//high mode on
                }
                else
                {
                    HIDBuffer[6] = 0;//high mode off
                    HIDBuffer[4] = 0xff;//low mode on
                }
                HID_Command(HIDBuffer, PS3Controller.OUTPUT_REPORT_BUFFER.Length + 2);
            }            
        }
        void hid_setLedOff(PS3Controller.LED a)
        {
            //check if LED is already off
            if ((byte)((byte)(((uint)a << 1) & HIDBuffer[11])) != 0)
            {
                //set the LED into the write buffer
                HIDBuffer[11] = (byte)((byte)(((uint)a & 0x0f) << 1) ^ HIDBuffer[11]);

                HID_Command(HIDBuffer, PS3Controller.OUTPUT_REPORT_BUFFER.Length + 2);
            }            
        }
        void hid_setLedOn(PS3Controller.LED a)
        {
            HIDBuffer[11] = (byte)((byte)(((uint)a & 0x0f) << 1) | HIDBuffer[11]);

            HID_Command(HIDBuffer, PS3Controller.OUTPUT_REPORT_BUFFER.Length + 2);            
        }
        void hid_enable_sixaxis()
        {            
            byte[] cmd_buf = new byte[12];
            cmd_buf[0] = 0x53;// HID BT Set_report (0x50) | Report Type (Feature 0x03)
            cmd_buf[1] = 0xF4;// Report ID
            cmd_buf[2] = 0x42;// Special PS3 Controller enable commands
            cmd_buf[3] = 0x03;
            cmd_buf[4] = 0x00;
            cmd_buf[5] = 0x00;

            HID_Command(cmd_buf, 6);                        
        }
        /************************************************************/
        /*                    L2CAP Commands                        */
        /************************************************************/
        void L2CAP_Command(byte[] data, int length)
        {
            byte[] buf = new byte[64];
            buf[0] = (byte)(hci_handle & 0xff);    // HCI handle with PB,BC flag
            buf[1] = (byte)(((hci_handle >> 8) & 0x0f) | 0x20);
            buf[2] = (byte)((4 + length) & 0xff);   // HCI ACL total data length
            buf[3] = (byte)((4 + length) >> 8);
            buf[4] = (byte)(length & 0xff);         // L2CAP header: Length
            buf[5] = (byte)(length >> 8);
            buf[6] = 0x01;  // L2CAP header: Channel ID
            buf[7] = 0x00;  // L2CAP Signalling channel over ACL-U logical link
            for (uint i = 0; i < length; i++)//L2CAP C-frame
            {
                buf[8 + i] = data[i];
            }
            BulkOutPipe.TransferData(buf, 0, length + 8);           
        }
        void l2cap_connection_response(byte rxid, byte[] dcid, byte[] scid, byte result)
        {            
            L2CAPBuffer[0] = L2CAP_CMD_CONNECTION_RESPONSE;// Code
            L2CAPBuffer[1] = rxid;// Identifier
            L2CAPBuffer[2] = 0x08;// Length
            L2CAPBuffer[3] = 0x00;
            L2CAPBuffer[4] = dcid[0];// Destination CID
            L2CAPBuffer[5] = dcid[1];
            L2CAPBuffer[6] = scid[0];// Source CID
            L2CAPBuffer[7] = scid[1];
            L2CAPBuffer[8] = result;// Result: Pending or Success
            L2CAPBuffer[9] = 0x00;
            L2CAPBuffer[10] = 0x00;//No further information
            L2CAPBuffer[11] = 0x00;

            L2CAP_Command(L2CAPBuffer, 12);            
        }        
        void l2cap_configure_request(byte rxid, byte[] dcid)
        {            
            L2CAPBuffer[0] = L2CAP_CMD_CONFIG_REQUEST;// Code
            L2CAPBuffer[1] = rxid;// Identifier
            L2CAPBuffer[2] = 0x08;// Length
            L2CAPBuffer[3] = 0x00;
            L2CAPBuffer[4] = dcid[0];// Destination CID
            L2CAPBuffer[5] = dcid[1];
            L2CAPBuffer[6] = 0x00;// Flags
            L2CAPBuffer[7] = 0x00;
            L2CAPBuffer[8] = 0x01;// Config Opt: type = MTU (Maximum Transmission Unit) - Hint
            L2CAPBuffer[9] = 0x02;// Config Opt: length            
            L2CAPBuffer[10] = 0xFF;// MTU
            L2CAPBuffer[11] = 0xFF;

            L2CAP_Command(L2CAPBuffer, 12);
        }
        void l2cap_config_response(byte rxid, byte[] scid)
        {            
            L2CAPBuffer[0] = L2CAP_CMD_CONFIG_RESPONSE;// Code
            L2CAPBuffer[1] = rxid;// Identifier
            L2CAPBuffer[2] = 0x0A;// Length
            L2CAPBuffer[3] = 0x00;
            L2CAPBuffer[4] = scid[0];// Source CID
            L2CAPBuffer[5] = scid[1];
            L2CAPBuffer[6] = 0x00;// Flag
            L2CAPBuffer[7] = 0x00;
            L2CAPBuffer[8] = 0x00;// Result
            L2CAPBuffer[9] = 0x00;
            L2CAPBuffer[10] = 0x01;// Config
            L2CAPBuffer[11] = 0x02;
            L2CAPBuffer[12] = 0xA0;
            L2CAPBuffer[13] = 0x02;

            L2CAP_Command(L2CAPBuffer, 14);
        }
        void l2cap_disconnection_request(byte rxid, byte[] dcid, byte[] scid)
        {
            L2CAPBuffer[0] = L2CAP_CMD_DISCONNECT_REQUEST;// Code
            L2CAPBuffer[1] = rxid;// Identifier
            L2CAPBuffer[2] = 0x04;// Length
            L2CAPBuffer[3] = 0x00;
            L2CAPBuffer[4] = scid[0];// Source CID
            L2CAPBuffer[5] = scid[1];
            L2CAPBuffer[4] = dcid[0];// Destination CID
            L2CAPBuffer[5] = dcid[1];

            L2CAP_Command(L2CAPBuffer, 6);
        }
        /************************************************************/
        /*                    HCI Commands                        */
        /************************************************************/
        void HCI_Command(byte[] buf, int nbytes)
        {
            hci_command_packets--;
            hci_event_flag &= ~(hci_event.HCI_FLAG_CMD_COMPLETE);
            //Host to device (0x00) | Class (0x20) | Device (0x00) - HCI_Command_Out, HCI_command_reg (0x00)
            //Single FunctionPrimary Controller - Standard request
            raw.SendSetupTransfer(0x20, 0x00, 0x0000, 0x0000, buf, 0, nbytes);
        }
        void hci_reset()
        {
            hci_event_flag = 0; // clear all flags
            HCIBuffer[0] = 0x03;// HCI OCF = 3
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x00; // parameter length = 0
            HCI_Command(HCIBuffer, 3);
            return;
        }
        void hci_read_buffer_size()
        {
            HCIBuffer[0] = 0x05; // HCI OCF = 5
            HCIBuffer[1] = 0x04 << 2; // HCI OGF = 4
            HCIBuffer[2] = 0x00;// parameter length = 0
            HCI_Command(HCIBuffer, 3);
            return;
        }
        void hci_read_local_version()
        {
            HCIBuffer[0] = 0x01; // HCI OCF = 1
            HCIBuffer[1] = 0x04 << 2; // HCI OGF = 4
            HCIBuffer[2] = 0x00;
            HCI_Command(HCIBuffer, 3);
            return;
        }
        void hci_write_scan_enable()
        {
            HCIBuffer[0] = 0x1A; // HCI OCF = 1A
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x01;// parameter length = 1
            HCIBuffer[3] = 0x02;// Inquiry Scan disabled. Page Scan enabled.
            HCI_Command(HCIBuffer, 4);
        }
        void hci_write_scan_disable()
        {
            HCIBuffer[0] = 0x1A; // HCI OCF = 1A
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x01;// parameter length = 1
            HCIBuffer[3] = 0x00;// Inquiry Scan disabled. Page Scan disabled.
            HCI_Command(HCIBuffer, 4);            
        }
        void hci_read_bdaddr()
        {
            HCIBuffer[0] = 0x09; // HCI OCF = 9
            HCIBuffer[1] = 0x04 << 2; // HCI OGF = 4
            HCIBuffer[2] = 0x00; // parameter length = 0
            HCI_Command(HCIBuffer, 3);
        }
        void hci_remote_name(byte disc_device)
        {            
            hci_event_flag &= ~(hci_event.HCI_FLAG_REMOTE_NAME);//Clear flag  
            HCIBuffer[0] = 0x19; // HCI OCF = 19
            HCIBuffer[1] = 0x01 << 2; // HCI OGF = 1
            HCIBuffer[2] = 0x0A; // parameter length 10
            HCIBuffer[3] = disc_bdaddr[0]; // 6 octet bdaddr
            HCIBuffer[4] = disc_bdaddr[1];
            HCIBuffer[5] = disc_bdaddr[2];
            HCIBuffer[6] = disc_bdaddr[3];
            HCIBuffer[7] = disc_bdaddr[4];
            HCIBuffer[8] = disc_bdaddr[5];
            HCIBuffer[9] = 0x01;//Page Scan Repetition Mode
            HCIBuffer[10] = 0x00;//Reserved
            HCIBuffer[11] = 0x00;//Clock offset - low byte
            HCIBuffer[12] = 0x00;//Clock offset - high byte
            HCI_Command(HCIBuffer, 13);
        }
        void hci_authentication_enable()
        {
            HCIBuffer[0] = 0x20; // HCI OCF = 20
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x01; // parameter length = 1
            HCIBuffer[3] = 0x00; // Authentication not required
            HCI_Command(HCIBuffer, 4);
        }

        void hci_accept_connection(byte disc_device)
        {            
            hci_event_flag &= ~(hci_event.HCI_FLAG_INCOMING_REQUEST);//Clear flag  
            HCIBuffer[0] = 0x09; // HCI OCF = 9
            HCIBuffer[1] = 0x01 << 2; // HCI OGF = 1
            HCIBuffer[2] = 0x07; // parameter length 7
            HCIBuffer[3] = disc_bdaddr[0]; // 6 octet bdaddr
            HCIBuffer[4] = disc_bdaddr[1];
            HCIBuffer[5] = disc_bdaddr[2];
            HCIBuffer[6] = disc_bdaddr[3];
            HCIBuffer[7] = disc_bdaddr[4];
            HCIBuffer[8] = disc_bdaddr[5];
            HCIBuffer[9] = 0x00; //switch role to master            
            HCI_Command(HCIBuffer, 10);
            return;
        }
        void hci_remote_version()
        {
            hci_event_flag &= ~(hci_event.HCI_FLAG_REMOTE_VERSION);//Clear flag
            HCIBuffer[0] = 0x1D; // HCI OCF = 1D
            HCIBuffer[1] = 0x01 << 2; // HCI OGF = 1
            HCIBuffer[2] = 0x02; // parameter length = 2
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCI_Command(HCIBuffer, 5);
        }
        void hci_change_packet_type()
        {
            hci_event_flag &= ~(hci_event.HCI_FLAG_CHANGE_PACKET_TYPE);//Clear flag
            HCIBuffer[0] = 0x0F; // HCI OCF = 0F
            HCIBuffer[1] = 0x01 << 2; // HCI OGF = 1
            HCIBuffer[2] = 0x04; // parameter length = 2
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCIBuffer[5] = 0x18;//DM1 and DH1 may be used
            HCIBuffer[6] = 0xCC;//DM3, DH3, DM5 and DH5 may be used
            HCI_Command(HCIBuffer, 7);

        }
        void hci_link_policy_settings()
        {            
            HCIBuffer[0] = 0x0D; // HCI OCF = 0D
            HCIBuffer[1] = 0x02 << 2; // HCI OGF = 2
            HCIBuffer[2] = 0x04; // parameter length = 2
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCIBuffer[5] = 0x0F;//Master/Slave Switch: Enabled, Hold Mode: Enabled, Sniff Mode: Enabled, Park Mode: Enabled
            HCIBuffer[6] = 0x00;
            HCI_Command(HCIBuffer, 7);                           	
        }
        void hci_link_supervision_timeout()
        {
            HCIBuffer[0] = 0x37; // HCI OCF = 37
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x04; // parameter length = 2
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCIBuffer[5] = 0x80;//10000 ms
            HCIBuffer[6] = 0x3E;
            HCI_Command(HCIBuffer, 7);
        }
        void hci_afh_host_channel_classification()
        {
            HCIBuffer[0] = 0x3F; // HCI OCF = 3F
            HCIBuffer[1] = 0x03 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x0A; // parameter length = 10
            HCIBuffer[3] = 0xFF;//AFH Mask
            HCIBuffer[4] = 0xFF;
            HCIBuffer[5] = 0xFF;
            HCIBuffer[6] = 0xFF;
            HCIBuffer[7] = 0x01;
            HCIBuffer[8] = 0x00;
            HCIBuffer[9] = 0x00;
            HCIBuffer[10] = 0xFC;
            HCIBuffer[11] = 0xFF;
            HCIBuffer[12] = 0x7F;            

            HCI_Command(HCIBuffer, 13); 
        }
        void hci_qos_setup()
        {
            hci_event_flag &= ~(hci_event.HCI_FLAG_QOS_SETUP_COMPLETE);
            HCIBuffer[0] = 0x07; // HCI OCF = 3F
            HCIBuffer[1] = 0x02 << 2; // HCI OGF = 3
            HCIBuffer[2] = 0x14; // parameter length = 20
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCIBuffer[5] = 0x00;// Flags - reserved for future use
            HCIBuffer[6] = 0x02;// Service type - Guaranteed

            HCIBuffer[7] = 0x00;// Token Rate
            HCIBuffer[8] = 0x00;
            HCIBuffer[9] = 0x00;
            HCIBuffer[10] = 0x00;

            HCIBuffer[11] = 0x00;// Peak Bandwidth
            HCIBuffer[12] = 0x00;
            HCIBuffer[13] = 0x00;
            HCIBuffer[14] = 0x00;

            HCIBuffer[15] = 0xF2;// Latency
            HCIBuffer[16] = 0x2B;
            HCIBuffer[17] = 0x00;
            HCIBuffer[18] = 0x00;

            HCIBuffer[19] = 0xFF;// Delay Variation:
            HCIBuffer[20] = 0xFF;
            HCIBuffer[21] = 0xFF;
            HCIBuffer[22] = 0xFF;

            HCI_Command(HCIBuffer, 23);
        }

        void hci_disconnect()
        {
            hci_event_flag &= ~(hci_event.HCI_FLAG_DISCONN_COMPLETE);
            HCIBuffer[0] = 0x06; // HCI OCF = 6
            HCIBuffer[1] = 0x01 << 2; // HCI OGF = 1
            HCIBuffer[2] = 0x03; // parameter length = 3
            HCIBuffer[3] = (byte)(hci_handle & 0xFF);//connection handle - low byte
            HCIBuffer[4] = (byte)((hci_handle >> 8) & 0x0F);//connection handle - high byte
            HCIBuffer[5] = 0x13; // reason = Remote User Terminated Connection
            HCI_Command(HCIBuffer, 6);
            return;
        }
        
        private bool hciflag(hci_event flag)
        {
            if ((hci_event_flag & flag) == flag)
                return true;
            else
                return false;
        }
        private bool l2capflag(l2cap_event flag)
        {
            if ((l2cap_event_status & flag) == flag)
                return true;
            else
                return false;
        }
        public static string ByteToHex(byte b)
        {
            const string hex = "0123456789ABCDEF";
            int low = b & 0x0f;
            int high = b >> 4;
            string s = new string(new char[] { hex[high], hex[low] });
            return s;
        }
        static void WriteSerial(string StringToWrite)
        {            
            // convert the string to bytes
            byte[] buffer = Encoding.UTF8.GetBytes(StringToWrite + "\r\n");            
            // send the bytes on the serial port
            UART.Write(buffer, 0, buffer.Length);
            //Debug.Print(StringToWrite);
        }
    }
}
