﻿using System.Text;
using System.IO.Ports;
using System.Threading;
using Microsoft.SPOT;
using GHIElectronics.NETMF.USBHost;

using PS3ControllerUSB;
using PS3ControllerBluetooth;

namespace PS3_Controller
{
    public class Program
    {       
        //PS3 Controllers
        const byte PS3Max = 4;//Max 4 controllers - this can be set to whatever you like
        static PS3Controller[] PS3 = new PS3Controller[PS3Max];
        static USBH_Device[] PS3Device = new USBH_Device[PS3Max];//Used to store the devices - see "DeviceDisconnectedEvent"                
        static byte PS3Connected;//Holds the numbers of controllers current connected

        //Bluetooth
        static Bluetooth BT;

        //Setup serial connection
        static SerialPort UART = new SerialPort("COM2", 115200);

        //Thread handlers
        static Thread PS3Handler = new Thread(PS3Thread);//Main Thread     
        static Thread RumbleHandler = new Thread(Rumble);//Handles the rumble function

        //Variables
        static bool[] runThread = new bool[PS3Max];//Set if loop should be running or not        
        static bool[] PS3Rumble = new bool[PS3Max];//Start rumble
        static bool[] print4DOF = new bool[PS3Max];//Print angles and gyro value

        //Used to store bluetooth address
        //static byte[] BDaddr = new byte[] { 0xC8, 0xBC, 0xC8, 0xB0, 0xB4, 0x71 };//My laptops bluetooth address
        
        public static void Main()
        {
            UART.Open();//Start serial communication on COM2
            WriteSerial("\r\nInitialized");//Indicate that it's up and running

            //Start USB Host events
            USBHostController.DeviceConnectedEvent += DeviceConnectedEvent;
            USBHostController.DeviceDisconnectedEvent += DeviceDisconnectedEvent;

            //Start threads            
            PS3Handler.Start();
            RumbleHandler.Start();

            //Sleep forever
            Thread.Sleep(Timeout.Infinite);
        }
        static void DeviceConnectedEvent(USBH_Device device)
        {            
            //Check if device is a PS3 Controller
            if (device.VENDOR_ID != PS3Controller.PS3_VENDOR_ID || device.PRODUCT_ID != PS3Controller.PS3_PRODUCT_ID)
            {
                if (device.TYPE == USBH_DeviceType.Hub)
                    WriteSerial("Hub Connected");
                else if (device.PRODUCT_ID == Bluetooth.CSR_PRODUCT_ID && device.VENDOR_ID == Bluetooth.CSR_VENDOR_ID)//Check if it is the bluetooth dongle
                {
                    WriteSerial("Bluetooth Dongle Connected");
                    BT = new Bluetooth(device);
                }
                else
                    WriteSerial("Unknown Device Connected");
                return;
            }
            if (PS3Connected == PS3Max)//Check if the maximum number of controller is exceeded
            {
                WriteSerial(PS3Max + " Controllers are already connected");
                return;
            }   

            PS3[PS3Connected] = new PS3Controller(device);//Connect the PS3 Controller
            PS3Device[PS3Connected] = device;//Store the device see "DeviceDisconnectedEvent"
            PS3SetLED(PS3Connected);//Set the specific LED on
            runThread[PS3Connected] = true;//Start the loop            
            WriteSerial("PS3 Controller: " + (PS3Connected + 1) + " - Connected");          

            PS3Connected++;                               
        }
        static void DeviceDisconnectedEvent(USBH_Device device)
        {
            //Check if device is a PS3 Controller
            if (device.VENDOR_ID != PS3Controller.PS3_VENDOR_ID || device.PRODUCT_ID != PS3Controller.PS3_PRODUCT_ID)
            {                
                if (device.TYPE == USBH_DeviceType.Hub)
                {
                    WriteSerial("Hub Disconnected");
                    for (byte i = 0; i < PS3Connected; i++)//All controllers has to be stopped
                    {
                        runThread[i] = false;
                        PS3[i].Abort();
                        BT.Abort();
                    }
                }
                else if (device.PRODUCT_ID == Bluetooth.CSR_PRODUCT_ID && device.VENDOR_ID == Bluetooth.CSR_VENDOR_ID)
                {
                    WriteSerial("Bluetooth Dongle Disconnected");
                    BT.Abort();
                }
                else
                    WriteSerial("Unknown Device Disconnected");
                return;
            }

            PS3Connected--;

            if (device.ID == PS3Device[PS3Connected].ID)//Check if it is the last one
            {                
                runThread[PS3Connected] = false;//Stop the loop for the last controller
                PS3[PS3Connected].Abort();//Abort the reading thread
                WriteSerial("PS3 Controller: " + (PS3Connected + 1) + " - Disconnected");
            }
            else
            {                
                byte PS3number;
                for (PS3number = 0; PS3number < PS3Connected; PS3number++)//Check which number that was disconnected
                    if (device.ID == PS3Device[PS3number].ID)
                        break;

                runThread[PS3number] = false;
                PS3[PS3number].Abort();
                WriteSerial("PS3 Controller: " + (PS3number + 1) + " - Disconnected");

                //Move all the controllers from that point one down
                for (byte i = PS3number; i < PS3Connected; i++)//Does not include the last one, as it is allways moved one down
                {                    
                    //Stop that thread including the next one
                    runThread[i] = false;
                    PS3[i].Abort();
                    runThread[i + 1] = false;
                    PS3[i + 1].Abort();

                    //Move one down
                    PS3[i] = new PS3Controller(PS3Device[i + 1]);
                    PS3Device[i] = PS3Device[i + 1];
                    PS3SetLED(i);

                    runThread[i] = true;
                    WriteSerial("PS3 Controller: " + (i + 2) + " - Changed to " + (i + 1));
                }             
            }
        }
        public static void PS3Thread()
        {
            string[] output = new string[PS3Max];
            string outputBT;
            string[] lastoutput = new string[PS3Max];
            string lastoutputBT = "";            
            while (true)
            {                
                if (Bluetooth.PS3BTConnected)
                {                    
                    outputBT = "";//Reset output
                    if(BT.GetButton(Bluetooth.Button.TRIANGLE))
                        outputBT += " - Triangle";
                    if(BT.GetButton(Bluetooth.Button.CIRCLE))
                        outputBT += " - Circle";
                    if (BT.GetButton(Bluetooth.Button.CROSS))
                        outputBT += " - Cross";
                    if (BT.GetButton(Bluetooth.Button.SQUARE))
                        outputBT += " - Square";

                    if (BT.GetButton(Bluetooth.Button.UP))
                        outputBT += " - Up";
                    if (BT.GetButton(Bluetooth.Button.RIGHT))
                        outputBT += " - Right";
                    if (BT.GetButton(Bluetooth.Button.DOWN))
                        outputBT += " - Down";
                    if (BT.GetButton(Bluetooth.Button.LEFT))
                        outputBT += " - Left";

                    if (BT.GetButton(Bluetooth.Button.L1))
                        outputBT += " - L1";
                    if (BT.GetButton(Bluetooth.Button.L2))
                        outputBT += " - L2";
                    if (BT.GetButton(Bluetooth.Button.L3))
                        outputBT += " - L3";
                    if (BT.GetButton(Bluetooth.Button.R1))
                        outputBT += " - R1";
                    if (BT.GetButton(Bluetooth.Button.R2))
                        outputBT += " - R2";
                    if (BT.GetButton(Bluetooth.Button.R3))
                        outputBT += " - R3";

                    if (BT.GetButton(Bluetooth.Button.SELECT))
                        outputBT += " - Select";
                    if (BT.GetButton(Bluetooth.Button.PS))
                        outputBT += " - PS";
                    if(BT.GetButton(Bluetooth.Button.START))
                        outputBT += " - Start";

                    //Ignore all joystick values if too small - 227 is center
                    if (BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatX) > 137 || BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatX) < 117)
                        outputBT += " - LeftHatX: " + BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatX);
                    if (BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatY) > 137 || BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatY) < 117)
                        outputBT += " - LeftHatY: " + BT.GetAnalogHat(Bluetooth.AnalogHat.LeftHatY);
                    if (BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatX) > 137 || BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatX) < 117)
                        outputBT += " - RightHatX: " + BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatX);
                    if (BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatY) > 137 || BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatY) < 117)
                        outputBT += " - RightHatY: " + BT.GetAnalogHat(Bluetooth.AnalogHat.RightHatY);

                    if (outputBT != "" && outputBT != lastoutputBT)//Check if output is not empty and not equal to the last one
                        WriteSerial("PS3 Controller" + outputBT);
                    lastoutputBT = outputBT;

                }
                
                for (byte i = 0; i < PS3Connected; i++)
                {
                    if (runThread[i])
                    {
                        output[i] = "";//Reset output

                        if (PS3[i].GetButton(PS3Controller.Button.TRIANGLE))
                            output[i] += " - Triangle";
                        if (PS3[i].GetButton(PS3Controller.Button.CIRCLE))
                            output[i] += " - Circle";
                        if (PS3[i].GetButton(PS3Controller.Button.CROSS))
                            output[i] += " - Cross";                        
                        if (PS3[i].GetButton(PS3Controller.Button.SQUARE))
                            output[i] += " - Square";

                        if (PS3[i].GetButton(PS3Controller.Button.UP))
                        {
                            output[i] += " - Up";
                            print4DOF[i] = true;//Start printing angles and gyro value
                        }
                        if (PS3[i].GetButton(PS3Controller.Button.RIGHT))
                        {
                            output[i] += " - Right";
                            PS3[i].SetBD_Addr(Bluetooth.BDaddr);//Set the bluetooth address into the controller
                            output[i] += " - Set BT Address: " + ByteToHex(Bluetooth.BDaddr[0]) + " " + ByteToHex(Bluetooth.BDaddr[1]) + " " + ByteToHex(Bluetooth.BDaddr[2]) + " " + ByteToHex(Bluetooth.BDaddr[3]) + " " + ByteToHex(Bluetooth.BDaddr[4]) + " " + ByteToHex(Bluetooth.BDaddr[5]);                            
                        }
                        if (PS3[i].GetButton(PS3Controller.Button.DOWN))
                        {
                            output[i] += " - Down";
                            print4DOF[i] = false;//Stop printing angles and gyro value
                        }
                        if (PS3[i].GetButton(PS3Controller.Button.LEFT))
                        {
                            output[i] += " - Left";
                            byte[] buffer = PS3[i].GetBD_Addr();//Read the bluetooth address in the controller
                            output[i] += " - Got BT Address: " + ByteToHex(buffer[0]) + " " + ByteToHex(buffer[1]) + " " + ByteToHex(buffer[2]) + " " + ByteToHex(buffer[3]) + " " + ByteToHex(buffer[4]) + " " + ByteToHex(buffer[5]);                            
                        }

                        if (PS3[i].GetButton(PS3Controller.Button.L1))
                            output[i] += " - L1";
                        if (PS3[i].GetButton(PS3Controller.Button.L2))
                            output[i] += " - L2: " + PS3[i].GetAnalogButton(PS3Controller.AnalogButton.L2);//Include analog readings - this can be done for almost all buttons
                        if (PS3[i].GetButton(PS3Controller.Button.L3))
                            output[i] += " - L3";
                        if (PS3[i].GetButton(PS3Controller.Button.R1))
                            output[i] += " - R1";
                        if (PS3[i].GetButton(PS3Controller.Button.R2))
                            output[i] += " - R2: " + PS3[i].GetAnalogButton(PS3Controller.AnalogButton.R2);//Include analog readings - this can be done for almost all buttons
                        if (PS3[i].GetButton(PS3Controller.Button.R3))
                            output[i] += " - R3";

                        if (PS3[i].GetButton(PS3Controller.Button.SELECT))
                        {
                            output[i] += " - Select";
                            output[i] += " - " + PS3[i].GetStatusString();//Print status string
                        }
                        if (PS3[i].GetButton(PS3Controller.Button.PS))//Reserved for bluetooth connection - still working on that
                            output[i] += " - PS";

                        //Ignore all joystick values if too small - 227 is center
                        if (PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatX) > 137 || PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatX) < 117)
                            output[i] += " - LeftHatX: " + PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatX);
                        if (PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatY) > 137 || PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatY) < 117)
                            output[i] += " - LeftHatY: " + PS3[i].GetAnalogHat(PS3Controller.AnalogHat.LeftHatY);
                        if (PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatX) > 137 || PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatX) < 117)
                            output[i] += " - RightHatX: " + PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatX);
                        if (PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatY) > 137 || PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatY) < 117)
                            output[i] += " - RightHatY: " + PS3[i].GetAnalogHat(PS3Controller.AnalogHat.RightHatY);

                        if (print4DOF[i])
                        {
                            //Print angles                   
                            output[i] += " - Pitch: " + PS3[i].GetAngle(PS3Controller.Angle.Pitch) + " - Roll: " + PS3[i].GetAngle(PS3Controller.Angle.Roll) + " - Gyro: " + PS3[i].GetSensor(PS3Controller.Sensor.gZ);
                        }

                        if (PS3[i].GetButton(PS3Controller.Button.START))
                        {
                            output[i] += " - Start";
                            PS3Rumble[i] = !PS3Rumble[i];//Set rumble
                            if (PS3Rumble[i])
                                output[i] += " - Rumble is on";
                            else
                                output[i] += " - Rumble is off";
                            while (PS3[i].GetButton(PS3Controller.Button.START)) ;//Wait for button to be released
                        }
                        
                        if (output[i] != "" && output[i] != lastoutput[i])//Check if output is not empty and not equal to the last one
                            WriteSerial("PS3 Controller: " + (i + 1) + output[i]);
                        lastoutput[i] = output[i];
                    }
                }
            }
        }

        public static void Rumble()
        {
            bool[] rumbleOff = new bool[PS3Max];
            while (true)
            {
                for (byte i = 0; i < PS3Connected; i++)
                {
                    if (PS3Rumble[i])
                    {
                        WriteSerial("PS3 Controller: " + (i + 1) + " - RumbleHigh");
                        PS3[i].SetRumbleOn(PS3Controller.Rumble.RumbleHigh);//Rumble with high intensity
                        rumbleOff[i] = true;
                    }
                    Thread.Sleep(2000 / PS3Connected);//2 seconds delay
                }                                
                for (byte i = 0; i < PS3Connected; i++)
                {
                    if (PS3Rumble[i])
                    {
                        WriteSerial("PS3 Controller: " + (i + 1) + " - RumbleLow");
                        PS3[i].SetRumbleOn(PS3Controller.Rumble.RumbleLow);//Rumble with low intensity                        
                        rumbleOff[i] = true;
                    }
                    Thread.Sleep(2000 / PS3Connected);//2 seconds delay
                }
                
                for (byte i = 0; i < PS3Connected; i++)
                {
                    if (!PS3Rumble[i] && rumbleOff[i])
                    {
                        WriteSerial("PS3 Controller: " + (i + 1) + " - RumbleOff");
                        PS3[i].SetRumbleOff();
                        rumbleOff[i] = false;
                    }
                }
            }
        }

        static void PS3SetLED(byte number)
        {
            PS3[number].SetLedOff(PS3Controller.LED.LED1);
            PS3[number].SetLedOff(PS3Controller.LED.LED2);
            PS3[number].SetLedOff(PS3Controller.LED.LED3);
            PS3[number].SetLedOff(PS3Controller.LED.LED4);

            if (number < 4)//Controller 1-4
                PS3[number].SetLedOn((PS3Controller.LED)(1 << number));

            else if (number < 7)//Controller 5-7
            {
                PS3[number].SetLedOn(PS3Controller.LED.LED4);
                PS3[number].SetLedOn((PS3Controller.LED)(1 << (number - 4)));
            }
            else if (number < 9)//Controller 8-9
            {
                PS3[number].SetLedOn(PS3Controller.LED.LED3);
                PS3[number].SetLedOn(PS3Controller.LED.LED4);
                PS3[number].SetLedOn((PS3Controller.LED)(1 << (number - 7)));
            }
            else//If you are lucky enough to have 10 controllers, it is supported
            {
                PS3[number].SetLedOn(PS3Controller.LED.LED1);
                PS3[number].SetLedOn(PS3Controller.LED.LED2);
                PS3[number].SetLedOn(PS3Controller.LED.LED3);
                PS3[number].SetLedOn(PS3Controller.LED.LED4);
            }
        }
        static void WriteSerial(string StringToWrite)
        {
            // convert the string to bytes
            byte[] buffer = Encoding.UTF8.GetBytes(StringToWrite + "\r\n");
            // send the bytes on the serial port
            UART.Write(buffer, 0, buffer.Length);
            Debug.Print(StringToWrite);//Also print for debugging
        }
        public static string ByteToHex(byte b)//Used to convert the bluetooth address to hex
        {
            const string hex = "0123456789ABCDEF";
            int low = b & 0x0f;
            int high = b >> 4;
            string s = "0x" + new string(new char[] { hex[high], hex[low] });
            return s;
        }
    }
}